\documentclass[structure=flat, language=pt]{cweb}

\usepackage[brazil]{babel}
\usepackage[latin1]{inputenc}
\usepackage{geometry}

\geometry{
verbose,
letterpaper,
tmargin=2cm,
bmargin=2cm,
lmargin=2cm,
rmargin=2cm,
headheight=0.5cm,
headsep=0.2cm,
footskip=0cm
}

\pagestyle{myheadings}
\markright{{\sc Estatsticas sobre rvores}}

\begin{document}

@* Introduo. O programa {\tt tree-stat} calcula estatsticas sobre vrios tipos
de rvores binrias. Ele pode ser utilizado para:

\begin{itemize}
\item Calcular a altura e o IPL mdios de rvores binrias geradas aleatriamente;
\item Criar figuras de rvores binrias geradas aleatriamente;
\item Calcular a altura e o IPL mdios de rvores de busca binrias geradas a partir
da insero de uma permutao qualquer de um conjunto de elementos;
\item Criar figuras de rvores binrias de busca geradas a partir do mtodo acima.
\end{itemize}

Os mdulos utilizados pelo programa so {\tt bin-tree}, que possui funes capazes
de gerar rvores binrias ao acaso, e {\tt bs-tree}, que possui funes capazes de 
manipular rvores binrias de busca.

@ {\it Linha de comando}. A linha de comando geral do programa  como segue:\medskip

\noindent {\tt tree-stat \{bp\} -[dch] -[srN]{\rm \it nmero} -[ltn]{\rm \it string}}\medskip

As opes {\tt b} ou {\tt p} so obrigatrias, devem comparecer logo aps o nome do
programa e apenas uma deve ser usada. A opo {\tt b} faz com que o programa compute
estatsticas sobre rvores binrias geradas ao acaso e a opo {\tt p} faz com que
as estatsticas sendo computadas sejam sobre rvores binrias de busca. As demais
opes esto detalhadamente explicadas abaixo.

\begin{description}
\item[{\tt d}.] Se presente, faz com que o programa gere apenas uma rvore e envie para
a sada padro o cdigo que a representa segundo a conveno do pacote PStricks,
disponvel para \LaTeX ou \TeX.
\item[{\tt h}.] Envia para sada padro dados sobre a distribuio da altura e do IPL
das rvores geradas. Os dados so dados na forma de uma tabela de frequncias.
\item[{\tt s}.] Deve ser seguida de um nmero, especificando a semente a ser utilizada
para iniciar o gerador de nmeros aleatrios.
\item[{\tt r}.] Especifica o nmero de experimentos que devem ser feitos para a determinao
das estatsticas. No tem efeito se utilizado com a opo {\tt -d}. O valor padro 
1000.
\item[{\tt N}.] Nmero de ns internos das rvores a serem geradas. Se no especificado, 
so geradas rvores com 100 ns internos.
\item[{\tt l}.] Tem efeito apenas se utilizado em conjunto com a opo {\tt -d}. Especifica
ento a separao entre dois nveis da rvore a ser gerada. Uma {\it string} como {\tt 2pt}
ou {\tt 2cm} deve vir em seguida. O padro  {\tt 5mm}.
\item[{\tt t}.] Como a anterior, s que especifica a distcia entre duas subrvores. O padro
 {\tt 5mm}.
\item[{\tt n}.] Tipo de n interno que ser utilizado. Apenas tem efeito em conjunto com a
opo {\tt -d}. O valor padro  \verb \Tc*{2pt} \ 
\footnote{Veja a documentao do pacote PStricks.}.
\item[{\tt c}.] Especifica que a sequncia dada para o tipo de n interno contm uma sequncia
de controle, que ser utilizada para imprimir o contedo do n de uma rvore binria de busca.
\end{description}

@ A estrutura geral do programa  a seguinte:

@c
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "bin-tree.h"
#include "bs-tree.h"
#define FALSE 0
#define TRUE !FALSE
@<Funo principal@>

@ @<Funo principal@>=
int main(int argc, char* argv[])
{
  @<Variveis locais@>@;
  @<Procura pelos argumentos de linha de comando@>@;
  @<Executa opes dadas na linha de comando@>@;
  return 0;
}

@ @<Procura pelos...@>=
{
  int i;

  if(argc < 2 || (argc >= 2 && (argv[1][0] != 'b' && argv[1][0] != 'p'))) {
    fprintf(stderr, "Uso: tree-stat {bp} -[dch] -[srN]nmero -[ltn]string\n");
    return 1;
  }
  
  for(i = 2; i < argc; i++) {
    if(argv[i][0] != '-') {
      fprintf(stderr, "Uso: tree-stat {b]} -[dch] -[srN]nmero -[ltn]string\n");
      return 1;
    }

    if(argv[i][1] == 'd') {
      draw_tree = TRUE;
      continue;
    }
    else if(argv[i][1] == 'h') {
      gen_tables = TRUE;
      continue;
    }
    else if(argv[i][1] == 'c') {
      control_seq = TRUE;
      continue;
    }
    else if(!argv[i][2]) continue;
    @#@#@#@;
    switch(argv[i][1]) {
    case 's':
      sscanf(argv[i] + 2, "%d", &seed);
      break;
    case 'r':
      sscanf(argv[i] + 2, "%d", &rep);
      break;
    case 'N':
      sscanf(argv[i] + 2, "%d", &N);
      break;
    case 'l':
      sscanf(argv[i] + 2, "%s", level_sep);
      break;
    case 't':
      sscanf(argv[i] + 2, "%s", tree_sep);
      break;
    case 'n':
      sscanf(argv[i] + 2, "%s", node_type);
      break;
    default:
      fprintf(stderr, "Uso: tree-stat {bp} -[dch] -[srN]nmero -[ltn]string\n");
      return 1;
    }
  }
}

@ Vamos agora definir as variveis locais que utilizamos na seo acima para guardar
os argumentos de linha de comando. Os valores de inicializao so os valores
padro.

@<Variveis locais@>=
int draw_tree = FALSE; /* Indica presena da opo {\tt -d} */
int control_seq = FALSE; /* Indica a presena da opo {\tt -c} */
int gen_tables = FALSE; /* Indica presena da opo {\tt -h} */
int seed = time(NULL); /* Semente do gerador */
int rep = 1000; /* Nmero de repeties dos experimentos */
int N = 100; /* Nmeros de ns internos das rvores */
char level_sep[] = "5mm\0         "; /* Separao entre os nveis */
char tree_sep[] = "5mm\0         "; /* Separao entre subrvores */
char node_type[] = "\\Tc*{2pt}\0         "; /* Estilo de n interno */

@ Alm dessas variveis, vamos utilizar duas outras que serviro para guardar
a altura e o IPL de uma rvore gerada.

@<Variveis locais@>+=
int height;
int ipl;

@ @<Executa opes dadas...@>=
srand(seed);
if(argv[1][0] == 'b') {
  if(!init_bin_tree_gen(N)) {
    fprintf(stderr, "Erro ao inicializar mdulo bin-tree\n");
    return 1;
  }
  if(draw_tree) {
    draw_random_tree(N, &height, &ipl, level_sep, tree_sep, node_type, stdout);
    fprintf(stderr, "Altura: %d\n", height);
    fprintf(stderr, "IPL: %d\n", ipl);
  }
  else @<Calcula estatsticas sobre rvores binrias@>@;
  deinit_bin_tree_gen();
}
else {
  bs_tree bst;
  if((bst = init_bst(N)) == NULL) {
    fprintf(stderr, "Erro ao inicializar mdulo bs-tree\n");
    return 1;
  }
  if(draw_tree) {
    gen_random_bst(bst);
    draw_bst(bst, control_seq, node_type, level_sep, tree_sep, stdout);
    fprintf(stderr, "Altura: %d\n", get_bst_height(bst));
    fprintf(stderr, "IPL: %d\n", get_bst_ipl(bst));
  }
  else @<Calcula estatsticas sobre rvores de busca binria@>@;
  destroy_bst(bst);
}

@* Estatsticas sobre rvores binrias. As estatsticas calculadas so basicamente
a altura e o IPL mdios. O EPL mdio tambm  calculado, a partir do IPL mdio, j
que ${\rm EPL} = {\rm IPL} + 2N$ para toda rvore binria com $N$ ns internos.

@<Calcula estatsticas sobre rvores binrias@>=
{
  if(gen_tables) {
    @<Inicializa vetores que guardam as tabelas de frequncia@>@;
    @<Gera |rep| rvores binrias e cria as tabelas@>@;
    @<Imprime as tabelas na sada padro e as elimina@>@;
  }
  else
    @<Gera |rep| rvores binrias e calcula seus parmetros mdios@>@;
}

@ Primeiramente, vamos tratar do caso em que no se deseja gerar as tabelas com a
distribuio de frequncias para os parmetros.

@<Gera |rep| rvores binrias e calcula seus...@>=
{
  int i;
  double ave_height = 0; /* Altura mdia */
  double ave_ipl = 0; /* IPL mdio */
  for(i = 0; i < rep; i++) {
    random_tree_params(N, &height, &ipl);
    ave_height += height;
    ave_ipl += ipl;
  }
  fprintf(stdout, "Altura mdia: %f\n", ave_height / rep);
  fprintf(stdout, "IPL mdio: %f\n", ave_ipl / rep);
  fprintf(stdout, "EPL mdio: %f\n", ave_ipl / rep + 2 * N);
}

@ Se a opo {\tt -h} foi utilizada, devemos gerar tabelas contendo as alturas das
rvores geradas e as frequncias com que ocorrem rvores com estas alturas, fazendo
algo semelhante para o IPL.

Isso pode consumir uma quantidade considervel de memria. Para uma rvore binria
qualquer, sabemos que

\[
N \lg \frac{N}{4} \leq {\rm IPL} \leq \frac{N (N-1)}{2}.
\]

Para uma rvore de 1000 ns internos, teramos ento $8000 \leq IPL \leq 499500$. 
Considerando que o espao que uma varivel do tipo |int| ocupa  de 4 bytes, a
tabela do IPL ocupara $1.90$mb em memria, incluindo as 8000 posies iniciais que
nunca sero utilizadas. Entretanto, estas posies ocupam apenas $31.25$kb. Portanto,
no vamos nos preocupar em economizar esta pequena quantidade de memria, o que
vai aumentar a velocidade dos clculos.

@<Inicializa vetores que guardam...@>=
int* height_table = NULL; /* Tabela das alturas */
int* ipl_table = NULL; /* Tabela dos IPLs */

if((height_table = calloc(N+1, sizeof(int))) == NULL) {
  fprintf(stderr, "Erro de alocao de memria\n");
  return 1;
}
if((ipl_table = calloc(1 + (N * (N-1)) / 2, sizeof(int))) == NULL) {
  fprintf(stderr, "Erro de alocao de memria\n");
  return 1;
}

@ @<Gera |rep| rvores binrias e cria as tabelas@>=
{
  int i;
  double ave_height = 0; /* Altura mdia */
  double ave_ipl = 0; /* IPL mdio */
  for(i = 0; i < rep; i++) {
    random_tree_params(N, &height, &ipl);
    ave_height += height;
    ave_ipl += ipl;
    height_table[height]++;
    ipl_table[ipl]++;
  }
  fprintf(stdout, "Altura mdia: %f\n", ave_height / rep);
  fprintf(stdout, "IPL mdio: %f\n", ave_ipl / rep);
  fprintf(stdout, "EPL mdio: %f\n", ave_ipl / rep + 2 * N);
}

@ Omitimos na impresso todas as entradas com frequncia ZERO, para facilitar a leitura
do resultado.

@<Imprime as tabelas...@>=
{
  int i, max;
  fprintf(stdout, "Tabela de alturas\n");
  for(i = 0, max = N+1; i < max; i++)
    if(height_table[i]) fprintf(stdout, "%d %d\n", i, height_table[i]);
  fprintf(stdout, "Tabela de IPLs\n");
  for(i = 0, max = 1 + (N * (N-1)) / 2; i < max; i++)
    if(ipl_table[i]) fprintf(stdout, "%d %d\n", i, ipl_table[i]);
  free(height_table);
  free(ipl_table);
}

@ Para calcularmos estatsticas sobre rvores de busca binria vamos proceder de 
maneira semelhante.

@<Calcula estatsticas sobre rvores de busca...@>=
{
  if(gen_tables) {
    @<Inicializa vetores que guardam as tabelas de frequncia@>@;
    @<Gera |rep| rvores de busca binria e cria as tabelas@>@;
    @<Imprime as tabelas na sada padro e as elimina@>@;
  }
  else
    @<Gera |rep| rvores de busca binria e calcula seus parmetros mdios@>@;
}

@ @<Gera |rep| rvores de busca binria e cria...@>=
{
  int i;
  double ave_height = 0; /* Altura mdia */
  double ave_ipl = 0; /* IPL mdio */
  for(i = 0; i < rep; i++) {
    gen_random_bst(bst);
    height = get_bst_height(bst);
    ipl = get_bst_ipl(bst);
    ave_height += height;
    ave_ipl += ipl;
    height_table[height]++;
    ipl_table[ipl]++;
  }
  fprintf(stdout, "Altura mdia: %f\n", ave_height / rep);
  fprintf(stdout, "IPL mdio: %f\n", ave_ipl / rep);
  fprintf(stdout, "EPL mdio: %f\n", ave_ipl / rep + 2 * N);
}

@ @<Gera |rep| rvores de busca binria e calcula seus...@>=
{
  int i;
  double ave_height = 0; /* Altura mdia */
  double ave_ipl = 0; /* IPL mdio */
  for(i = 0; i < rep; i++) {
    gen_random_bst(bst);
    ave_height += get_bst_height(bst);
    ave_ipl += get_bst_ipl(bst);
  }
  fprintf(stdout, "Altura mdia: %f\n", ave_height / rep);
  fprintf(stdout, "IPL mdio: %f\n", ave_ipl / rep);
  fprintf(stdout, "EPL mdio: %f\n", ave_ipl / rep + 2 * N);
}

@
\end{document}
\fi
\inx
\fin
\con
