/*
 * This code is from "Algorithms in C, Third Edition,"
 * by Robert Sedgewick, Addison Wesley Longman, 1998.
 * (Mescla de prog12.5.c e prog14.3.c)
 */
#include "Item.h"
#include "ST.h"

typedef struct STnode *link;
struct STnode { Item item; link next; };

static link *heads, z;
static int N, M;

static link NEW(Item item, link next)
{
    link x = malloc(sizeof *x);
    x->item = item;
    x->next = next;
    return x;
}

int STcount()
{
    return N;
}

/***********************************************************************
 *
 * Funcao de hashing universal para chaves do tipo string
 *
 ***********************************************************************/

int hashU(char *v, int M)
  { int h, a = 31415, b = 27183;
    for (h = 0; *v != '\0'; v++, a = a*b % (M-1)) 
        h = (a*h + *v) % M;
    return h;
  }

/***********************************************************************
 *
 * Outra funcao de hashing; devida a Knuth (aparece no Stanford GraphBase) 
 *
 * http://www-cs-staff.Stanford.EDU/~knuth/sgb.html
 *
 ***********************************************************************/

#define HASH_MULT 314159 /* random multiplier */
#define HASH_PRIME 516595003 /* the 27182818th prime; it's less than $2^{29}$ */

long hash_Knuth(char *t)
{ register long h;
  for (h=0; *t; t++) {
    h+=(h^(h>>1))+HASH_MULT*(unsigned char)*t;
    while (h>=HASH_PRIME) h-=HASH_PRIME;
  }
  return h;
}

/***********************************************************************/

#define hash(v,M) (hash_Knuth(v)%M)

void STinit(int max) 
  { int i;
    N = 0; M = max/5;
    heads = malloc(M*sizeof(link));
    z = NEW(NULLitem, NULL);
    for (i = 0; i < M; i++) heads[i] = z; 
  }

Item searchR(link t, Key v)
  { 
    if (t == z) return NULLitem;
    if (eq(key(t->item), v)) return t->item;
    return searchR(t->next, v);
  }

Item STsearch(Key v)
  { return searchR(heads[hash(v, M)], v); }

void STinsert(Item item)
  { int i = hash(key(item), M);
    heads[i] = NEW(item, heads[i]); N++; }

/*
 * void STdelete(Item item)
 *   { int i = hash(key(item), M);
 *     heads[i] = deleteR(heads[i], item); }
 */
/***********************************************************************/

void STshow_all()
{
    int i, no_per_list, max_list = -1;
    link t;

    for (i=0; i<M; i++) {
      if (heads[i] != z)
	printf("*** list %d:\n", i);
      for (no_per_list=0, t = heads[i]; t != z; no_per_list++, t = t->next) 
	ITEMshow(t->item);
      if (no_per_list>max_list) max_list=no_per_list;
    }

    printf("\nMax no. of items in the lists: %d\n", max_list);
    printf("Average no. of items in the lists: %g\n\n", (double)N/M);
}

