/*
 * This code is from "Algorithms in C, Third Edition,"
 * by Robert Sedgewick, Addison Wesley Longman, 1998.
 * (with small alterations...)
 */
#include <stdlib.h>
#include "Item.h"
#define o ++n_probes
#define null(A) (o, st[A] == NULLitem)
static int N, M;
static long n_probes;
static Item *st;

/***********************************************************************
 *
 * Funcao de hashing; devida a Knuth (aparece no Stanford GraphBase) 
 *
 * http://www-cs-staff.Stanford.EDU/~knuth/sgb.html
 *
 ***********************************************************************/

#define HASH_MULT 314159 /* random multiplier */
#define HASH_PRIME 516595003 /* the 27182818th prime; it's less than $2^{29}$ */

long hash_Knuth(char *t)
{ register long h;
  for (h=0; *t; t++) {
    h+=(h^(h>>1))+HASH_MULT*(unsigned char)*t;
    while (h>=HASH_PRIME) h-=HASH_PRIME;
  }
  return h;
}

/***********************************************************************/

#define hash(v,M) (hash_Knuth(v)%M)

/***********************************************************************/

void STinit()
  { int i; 
    N = 0; M = 4;
    st = malloc(M*sizeof(Item));
    for (i = 0; i < M; i++) st[i] = NULLitem; 
  }
int STcount() { return N; }

/***********************************************************************
 * Programa 14.7
 ***********************************************************************/

void init(int M)
  { int i; N = 0;
    st = malloc(M*sizeof(Item));
    for (i = 0; i < M; i++) st[i] = NULLitem; 
  }

void expand();
void STinsert(Item item)
  { Key v = key(item);
    int i = hash(v, M);
    while (!null(i)) i = (i+1) % M;
    st[i] = item; 
    if (N++ > M/2) expand();
  }
void expand()
  { int i; Item *t = st;
    M=M+M;
    init(M);
    for (i = 0; i < M/2; i++) 
      if (o, t[i] != NULLitem)
        STinsert(t[i]); 
    free(t);
  }

Item STsearch(Key v)
  { int i = hash(v, M);
    while (!null(i))
      if eq(v, key(st[i])) return st[i]; 
      else i = (i+1) % M;
    return NULLitem;
  }

/***********************************************************************/

void STshow_all()
{
    int i;

    for (i=0; i<M; i++)
      if (!null(i))
	ITEMshow(st[i]);

    printf("\nFinal table load factor: %g\n", (double)N/M);
    printf("Number of probes: %ld\n", n_probes - M);
    printf("Average number of probes per insertion: %g\n\n",
            (double)(n_probes-M)/N);
    /* We probe each entry of the table for printing; hence `- M'. */
}

/***********************************************************************/
int compare(const void *i, const void *j)
{
  if (num(*(Item *)i)>num(*(Item *)j))
    return -1;
  else if (num(*(Item *)i)<num(*(Item *)j))
    return 1;
  else if (less(key(*(Item *)i),key(*(Item *)j)))
    return -1;
  else
    return 1;
}

void STsort(void (*visit)(Item))
{
  Item *sorted_t = malloc(N*sizeof(Item));
  int i, j;

  for (i=0,j=0; j<M; j++)
    if (!null(j))
      sorted_t[i++]=st[j];

  qsort(sorted_t, N, sizeof(Item), compare);

  for (i=0; i<N; i++)
    ITEMshow(sorted_t[i]);    
}
