/*
 * This code is from "Algorithms in C, Third Edition,"
 * by Robert Sedgewick, Addison Wesley Longman, 1998.
 * (Com pequenas alteracoes).
 */
#include <stdlib.h>
#include "Item.h"
typedef struct STnode* link;
struct STnode { Item item; link l, r; int N; int red; };
static link head, z;
link NEW(Item item, link l, link r, int N, int red)
  { link x = malloc(sizeof *x); 
    x->item = item; x->l = l; x->r = r; x->N = N; x->red = red;
    return x;
  }
void STinit()
  { head = (z = NEW(NULLitem, 0, 0, 0, 0)); }
int STcount() { return head->N; }
Item searchR(link h, Key v)
  { Key t;
    if (h == z) return NULLitem;
    t = key(h->item);
    if eq(v, t) return h->item;
    if less(v, t) return searchR(h->l, v);
             else return searchR(h->r, v);
  }
Item STsearch(Key v) 
  { return searchR(head, v); } 

/* prog12.11 - Rotations */

link rotR(link h)
  { link x = h->l; h->l = x->r; x->r = h;
    h->N = h->l->N + h->r->N + 1;
    x->N = x->l->N + x->r->N + 1;
    return x; }

link rotL(link h)
  { link x = h->r; h->r = x->l; x->l = h;
    h->N = h->l->N + h->r->N + 1;
    x->N = x->l->N + x->r->N + 1;
    return x; }

/* Implementacao de fixN() */

void fixN(link h)
{ if (h != z)
  { fixN(h->l); fixN(h->r); h->N = h->l->N + h->r->N +1; }
}  

/* prog13.6 - Insercao em ABBs rubro-negras */ 

#define hl  (h->l)
#define hr  (h->r)
#define hll  (h->l->l)
#define hlr  (h->l->r)
#define hrl  (h->r->l)
#define hrr  (h->r->r)

link RBinsert(link h, Item item, int sw)
  { Key v = key(item);
    if (h == z) return NEW(item, z, z, 1, 1);  
    if ((hl->red) && (hr->red)) 
      { h->red = 1; hl->red = 0; hr->red = 0; }
    if (less(v, key(h->item))) 
      { 
        hl = RBinsert(hl, item, 0); 
        if (h->red && hl->red && sw) h = rotR(h); 
        if (hl->red && hll->red) 
          { h = rotR(h); h->red = 0; hr->red = 1; }
      }
    else
      { 
        hr = RBinsert(hr, item, 1); 
        if (h->red && hr->red && !sw) h = rotL(h); 
        if (hr->red && hrr->red) 
          { h = rotL(h); h->red = 0; hl->red = 1; }
      }
    fixN(h); return h;
  }

void STinsert(Item item)
  { head = RBinsert(head, item, 0); head->red = 0; }

/* Rotinas de impressao */

void print_in_orderR(link h)
{
  if (h!=z) {
    print_in_orderR(h->l);
    ITEMshow(h->item);
    print_in_orderR(h->r);
  }
}

void STshow_all()
{
  print_in_orderR(head);
}

/* prog12.13 - Selecao em uma ABB */

Item selectR(link h, int k)
  { int t = h->l->N;
    if (h == z) return NULLitem;
    if (t > k) return selectR(h->l, k);
    if (t < k) return selectR(h->r, k-t-1);
    return h->item;
  }
Item STselect(int k)
  { return selectR(head, k); } 

/* prog12.14 - Selecao em uma ABB, levando o elemento ate a raiz */

link partR(link h, int k)
  { int t = h->l->N; 
    if (t > k )
      { h->l = partR(h->l, k); h = rotR(h); }
    if (t < k )
      { h->r = partR(h->r, k-t-1); h = rotL(h); }
    return h;
  }

/* prog13.1 - Balanceamento explicito de uma ABB */

link balanceR(link h)
  { 
    if (h->N < 2) return h;
    h = partR(h, h->N/2);
    h->l = balanceR(h->l); 
    h->r = balanceR(h->r);
    return h;
  }

void BSTbalance()
  { head = balanceR(head); }

/* prog5.18 - Impressao de uma arvore binaria */

void printnode(Item item, int h)
  { int i;
    for (i = 0; i < h; i++) printf("   ");
    ITEMshow(item);
  }

void show_in(link x, int h)
  { 
    if (x == z) { printnode(NULLitem, h); return; }
    show_in(x->r, h+1);    
    printnode(x->item, h);
    show_in(x->l, h+1);    
  }

void BSTquickprint()
  { show_in(head, 0); }
